package Xchat::b0at::thermodynambot;
# 
# No modifications to the code are necessary to use it.
# 
use strict;
use warnings;
use POSIX 'strftime';
{ # manually "import" Xchat:: functions for easy use later
  # "use Xchat 'foo';"-style importing only works in 2.4.0+
  # sub/const name is exactly the same unless specified otherwise
	*register       = \&Xchat::register;
	*hook_server    = \&Xchat::hook_server;
	*hook_command   = \&Xchat::hook_command;
	*hook_print     = \&Xchat::hook_print;
	*hook_timer     = \&Xchat::hook_timer;
	*unhook         = \&Xchat::unhook;
	*xprint         = \&Xchat::print;  # name differs
	*xprintf        = \&Xchat::printf; # name differs
	*command        = \&Xchat::command;
	*commandf       = \&Xchat::commandf;
	*find_context   = \&Xchat::find_context;
	*get_context    = \&Xchat::get_context;
	*set_context    = \&Xchat::set_context;
	*get_info       = \&Xchat::get_info;
	*get_prefs      = \&Xchat::get_prefs;
	*emit_print     = \&Xchat::emit_print;
	*nickcmp        = \&Xchat::nickcmp;
	*get_list       = \&Xchat::get_list;
	*user_info      = \&Xchat::user_info;
	*context_info   = \&Xchat::context_info;
	sub EAT_NONE   (){Xchat::EAT_NONE}
	sub EAT_XCHAT  (){Xchat::EAT_XCHAT}
	sub EAT_PLUGIN (){Xchat::EAT_PLUGIN}
	sub EAT_ALL    (){Xchat::EAT_ALL}
	sub REMOVE     (){0} # for timers for all xchat versions
	sub KEEP       (){1}
}
my $NAME    = 'thermodynambot';
my $VERSION = '004-01';
my $PREFIX  = "\02TDB\02\t";
register($NAME, $VERSION, 'The reaction script.');

use constant CONF_FILE    => '_thermodynambot.conf'; # filename looked for in 'xchatdir'
use constant MACRO_CHAR   => '%'; # for conf macros
use constant SEPARATOR    => ':'; # for conf properties
use constant NO_MODE      => 'none'; # mode string for no-mode user
use constant REGEX_FLAG   => '&'; # something that can't start a user mask
use constant NETWORK_SEP  => ':'; # network:#channel in +-in/where

my $patterns; # script-wide config entries stored here
my %COMMANDS = (
	'THERMODYNAMBOT' => [ sub{ hookage(); return EAT_ALL; }, 'Usage: THERMODYNAMBOT, reloads '.CONF_FILE ],
	'TDB'            => [ sub{ hookage(); return EAT_ALL; }, 'Usage: TDB'.        ', reloads '.CONF_FILE ],
	'TDBECHO' => [ \&user_test,
		" <nick!user\@addr> '<mode>' <ispriv> <isaction> <network or server> <message text>\n".
		"	   quotes on mode are required (eg '\@', '+', '')\n".
		"	   ispriv and isaction should be 1 or 0\n".
		"	   TDBTEST runs the commands, TDBECHO prints the commands instead\n" ],
);
$COMMANDS{'TDBTEST'} = [ \&user_test, 'Usage: TDBTEST'.$COMMANDS{'TDBECHO'}->[1] ];
$COMMANDS{'TDBECHO'}->[1] = 'Usage: TDBECHO'.$COMMANDS{'TDBECHO'}->[1];
hook_command( $_, $COMMANDS{$_}->[0], { help_text=>$COMMANDS{$_}->[1] } ) for( keys %COMMANDS );
xprint( "\02$NAME v$VERSION\02 by b0at (use /".(join(', /',keys %COMMANDS)).")" ); # announce

my @privmsg_hooks = (); # for hooking/unhooking events
hookage(1); # do config and hook events for the first time

# handle differences in events, then pass to respond, then to do_commands to execute
# stripping color codes from the nick is key
sub hdl_chan       {
	my ($nick, $text, $mode) = @{$_[0]};
	if( my $cmds = respond({nick=>strip_codes($nick), msg=>$text, mode=>$mode, private=>0, action=>0}) )
	{ do_commands( $cmds, get_context() ) }
	return EAT_NONE;
}
sub hdl_actn       {
	my ($nick, $text) = @{$_[0]};
	if( my $cmds = respond({nick=>strip_codes($nick), msg=>$text, mode=>undef, private=>0, action=>1}) )
	{ do_commands( $cmds, get_context() ) }
	return EAT_NONE;
}
sub hdl_priv       {
	my ($nick, $text) = @{$_[0]};
	if( my $cmds = respond({nick=>strip_codes($nick), msg=>$text, mode=>'',    private=>1, action=>0}) )
	{ do_commands( $cmds, get_context() ) }
	return EAT_NONE;
}
sub user_test      { # handle user command triggering for testing
	my ($word, $eol) = @_;
	if( $$word[6] and $$word[2]=~/^(['"])(.*)\1$/ ) {
		my ($nick,$host) = split '!', $$word[1];
		my $cmds = respond( 
		{
			msg=>$$eol[6], 
			#mode=>$2, 
			private=>$$word[3], 
			action=>$$word[4], 
			network=>$$word[5],
		}, 
		{
			nick=>$nick, 
			host=>$host, 
			prefix=>$2
		});
		if( uc $$word[0] eq 'TDBECHO' ) { xprint("$PREFIX $_") foreach( @$cmds ) }
		else { do_commands($_, get_context()) foreach( @$cmds )	}
	}
	else { xprint( $COMMANDS{TDBTEST}->[1] ) }
	return EAT_ALL; # don't pass to server
}
# do things
sub respond        { # decide what to do based on event and entries
	# This sub does one thing and does it well: causes heartburn.
	my ($event, $user) = @_; # info hashes
	# event: nick, msg, private (bool), action (bool), [mode, network, where, context]
	# user: (from user_info or /who output) nick, host, prefix, away, [user, addr, who]
	
	# clarify event data
	$$event{where}     = get_info('channel'); # tab name
	$$event{network} ||= get_info('network') || get_info('server') || 'offline';

	# get user info
	$user = user_info($$event{nick}) if not ($$user{host} or $$user{nick});
	if( not $$user{host} ) {
		# hook and send a /who instead 
		# ### maybe make it search through all contexts first later
		# (use $$event{nick} here because we haven't gotten user_info yet)
		my ($context, $hook352, $hook315) = (get_context(), );
		$hook352 = hook_server('352', sub { # hook /who output
			my ($w,$e)=@_;
			return EAT_NONE if(nickcmp($$w[7],$$event{nick})!=0);
			my $cmds = respond($event, {
				away => ($$w[8]=~/G/?1:0),
				nick => $$w[7], # actual, case-correct nick
				host => $$w[4].'@'.$$w[5],
				prefix => $$event{mode}||'', # doesn't apply if they're not in the channel
				# for convenience
				user => $$w[4], addr => $$w[5],
				who  => $$w[7].'!'.$$w[4].'@'.$$w[5]
			});
			unhook($hook352);
			undef($hook352);
			do_commands( $cmds, $context );
			return EAT_XCHAT; # hide our /who from user
		});
		$hook315 = hook_server('315', sub { # hide "End of /who output" line
			return EAT_NONE if $hook352;
			unhook($hook315); return EAT_XCHAT;
		});
		command("who $$event{nick}");
		return; # exit sub and do callback when we get /who output
	}
	
	# for convenience
	($$user{user}, $$user{addr}) = split '@', $$user{host}
		if not $$user{user} and not $$user{addr};
	$$user{who} = "$$user{nick}!$$user{host}" if not $$user{who};

	my @commands;
	
	# check patterns against event data
PATTERN:for my $entry (@$patterns)
	{
		# text
		next if not my @match = $$event{msg}=~/$entry->{pattern}/;
#		next if not my @match = $$event{msg}=~/^$entry->{pattern}/; # ?
		
		my $check;
		
		for my $type ('action', 'private') {
			if( $$event{$type} ) {
				# excluded networks
				$$event{network}=~/^$_$/ and next PATTERN for ( @{$$entry{$type}[0]} );
				
				# allowed networks
				$check = 0;
				$$event{network}=~/^$_$/ and ++$check and last for ( @{$$entry{$type}[1]} );
				next PATTERN if not $check;
			}
		}

		# public events
		if( not $$event{private} ) {
			# excluded network:channel
			for ( @{$$entry{network}[0]} )
			{
				next PATTERN if( 
					  $$event{network} =~ /^$_->[0]$/    # network disallowed
					and $$event{where} =~ /^$_->[1]$/ ); # channel on network disallowed
			}
			# excluded channel
			$$event{where}=~/^$_$/ and next PATTERN for ( @{$$entry{where}[0]} );
			
			# allowed channels
			$check= 0 ;
			$$event{where}=~/^$_$/ and ++$check and last for ( @{$$entry{where}[1]} );
			if ( not $check )
			{	# if not allowed yet by plain channels, check allowed network/channel combos
				for ( @{$$entry{network}[1]} )
				{
					++$check and last if( 
						  $$event{network} =~ /^$_->[0]$/    # network allowed
						and $$event{where} =~ /^$_->[1]$/ ); # channel on network allowed
				}
			}
			next PATTERN if not $check;
			
			# excluded modes (optional)
			next if not $$user{prefix} # user has no mode
				and defined $$entry{nomode}[0]; # excluding (0) no-moders
			
			$$user{prefix} eq $_ and next PATTERN for( @{$$entry{mode}[0]} );
			
			# allowed modes (optional)
			if( scalar @{$$entry{mode}[1]} or defined $$entry{nomode}[1] ) {
				$check = 0;
				++$check if defined $$entry{nomode}[1] and not $$user{prefix}; ####?
				if( not $check ) {
					$$user{prefix} eq $_ and ++$check and last
						for ( @{$$entry{mode}[1]} );
					next PATTERN if not $check; # restricted to these modes only
				}
			}
		} # public
		# excluded people
		$$user{who}=~/^$_$/ and next PATTERN for ( @{$$entry{from}[0]} );

		# allowed people
		$check = 0;
		$$user{who}=~/^$_$/ and ++$check and last for ( @{$$entry{from}[1]} );
		next if not $check;

		# --- event matches, push commands ---
		push @commands, expand_macros( $$entry{commands}, \@match, $event, $user);
	} # PATTERN

	return \@commands;
}
sub do_commands    {
	my ($cmds, $context) = @_;
	hook_timer(0, # immediately after this event
	sub {
		#my $original_context = get_context(); # uncomment for polite context switching
		set_context( $context );
		command( $cmds );
		#set_context($original_context); # uncomment for polite context switching
		return REMOVE; # do this timer only once
	} );
}
sub load_conf($)   { # read in a conf file, return valid data, print on error, stop on fatal error
	my ($filename, # only argument
		$count, $fh, $nc) = (shift, 0,); # nc==new conf
	return unless( $filename and -e $filename and open $fh, $filename );

	my @p; # store patterns here

	while( my $line = <$fh> )
	{
		++$count; # line count
		next if $line=~/^\s*$|^#/;
		last if $line=~/^__END__\s*$/;
		
		if( my ($method, $pattern) = $line=~/^\s*\[(text|regex)\s*(??{SEPARATOR})\s*(.*?)\]\s*$/i )
		{
			# add a new entry
			push @p, 
				{ # criteria => [ negative (0), positive (1) ],
					from     => [ [],			[]		  ], # people
					where    => [ [],			[]		  ], # places
					network  => [ [],			[]		  ], # places w/network specified
					mode     => [ [],			[]		  ], # status
					# special values, bool
					nomode   => [ undef,		 undef	   ], # probably a better way, but it's fine for now
					#			networks allowed, networks disallowed
					private  => [ [],			[]		  ],
					action   => [ [],			[]		  ],
					# just a list
					commands => [],							 # results
				};
			
			if( lc $method eq 'text' )
			{   # (?i) makes it case insensitive
				$p[$#p]{pattern} = '(?i)'.wildcard($pattern);
			}
			elsif( lc $method eq 'regex' )
			{
				$p[$#p]{pattern} = $pattern;
			}
		}
		# bad form, don't risk overlapping entries' data
		elsif ( $line=~/^\s*\[.*\]\s*$/i )
		{
			xprint("$PREFIX Parse error: '$_' is not of the form ",
				"'[regex:REGULAR EXPRESSION]' or '[text:WILDCARD EXPRESSION]',");
			xprint("$PREFIX \cBParse halted.\cB");
			last;
		}
		
		elsif( my ($where, $wherelist) = $line=~/^\s*(\+|-)(?:in|where)\s*(??{SEPARATOR})\s*(.*)$/i )
		{
			$where = ($where eq '+') ? 1 : 0;
			
			# put the rest into +-where or +-network
			for my $place ( split(/\s*,\s*/, $wherelist) )
			{
				my ($net, $type, $chan);
				
				if( ($net, $type) = $place=~/^(?:(.+?)(??{NETWORK_SEP}))?(action|private)$/i )
				{
					if( not $net )
					{  # 'private' or 'action' without a network
						# set allowed/disallowed on all networks
						push @{ $p[$#p]{ lc $type }[$where] }, '(?i)'.wildcard('*'); 
					}
					else
					{ # network:private or network:action
						push @{ $p[$#p]{ lc $type }[$where] }, '(?i)'.wildcard($net);
					}
				}			
				# network specified, de-wildcard and stick in an array red
				elsif( ($net, $chan) = $place=~/^(.+?)(??{NETWORK_SEP})(.+?)$/ )
				{ ### unique structure
					push @{ $p[$#p]{network}[$where] },
						[ '(?i)'.wildcard($net), '(?i)'.wildcard($chan) ];
				}
				else # regular channel wildcard
				{
					push @{ $p[$#p]{where}[$where] }, '(?i)'.wildcard($place);
				}
			}
		}
		
		elsif( my ($from, $fromlist) = $line=~/^\s*(\+|-)from\s*(??{SEPARATOR})\s*(.*)$/i )
		{
			$from = ($from eq '+') ? 1 : 0;
			
			# no special values, go straight into parsing the list
			for( split /\s*,\s*/, $fromlist )
			{
				# full user mask
				if( /^(??{REGEX_FLAG})([^!]+![^@]+@.+)$/ )
				{
					push @{ $p[$#p]{from}[$from] }, $2;
				}
				elsif( /^[^!]+![^@]+@.+$/ )
				{
					push @{ $p[$#p]{from}[$from] }, wildcard($_); # <003-02> was $2 instead of $_, which was totally wrong
				}
				
				# lonesome nick: bob!
				elsif( /^[^!]+!$/ )
				{
					push @{ $p[$#p]{from}[$from] }, wildcard($_).".*?@.*?";
				}
				
				# lonesome address: @foo.com
				elsif( /^\@.+$/ )
				{
					push @{ $p[$#p]{from}[$from] }, ".*?!.*?".wildcard($_);
				}
				
				# unknown -- assume a wildcard expression, eg *
				else 
				{
					push @{ $p[$#p]{from}[$from] }, wildcard($_);
					# maybe issue a warning?
					#xprint("$PREFIX Parse warning line $count: '$_' ",
					#   "interpreted as a wildcard expression"); 
				}
			}
		}
		
		elsif( my ($mode, $modelist) = $line=~/^\s*(\+|-)mode\s*(??{SEPARATOR})\s*(.*)$/i )
		{
			$mode = ($mode eq '+') ? 1 : 0;
			
			# special value: NO_MODE (constant)
			my %modelist; @modelist{ split /\s*,\s*/, $modelist } = undef;
			if( exists $modelist{NO_MODE.''} ) # 003-03: ugh, interpreted NO_MODE as a string instead of a constant
			{
				$p[$#p]{nomode}[$mode] = $mode; # set idx 1 or 0 to a defined value
				delete $modelist{NO_MODE.''}; # don't include in list below
			}
			
			push @{ $p[$#p]{mode}[$mode] }, keys %modelist;
		}
		
		elsif( my ($command) = $line=~/^\s*command\s*(??{SEPARATOR})\s*(.*)$/i ) 
		{
			push @{ $p[$#p]{commands} }, $command;
		}
		
		else
		{
			xprint("$PREFIX Parse error on line $count: unknown data.");
		}
	}
	close $fh;

	return \@p;
}
# helpers
sub wildcard($)    { # convert wildcard pattern to capturing regex for matching
	local $_ = shift;
	   
	# special characters allowed in nicks/channels
	s!([\$\%\@\\\][(){}|^"`'.])!\\$1!g;
	
	# capture these
	s!\?!(.)!g;   # one character
	s!\*(.+)!(.*?)$1!g; # non-greedy is required here
	s!\*$!(.*)!g; # greedy is required here
	
	# make spaces flexible?
	#s!\s+!\\s\+!g;

	return $_;
}
sub expand_macros  { # expand macros in commands based on event and user data
	my @cmds = @{ shift @_ }; # make a copy or will modify pattern data
	my ($matches, $event, $user) = @_;

	for my $cmd (@cmds) {
		# split on macro pattern (captured), replace as needed
		my @bits = split(/( (??{MACRO_CHAR}) . )/x, $cmd);
		for my $bit (@bits) {
SWITCH:	for( $bit=~/^(??{MACRO_CHAR})(.)$/ ) { # for %_ bits
				# user data: n, u, a, c, s, w, o
				/n/ and do { $bit = $$user{nick};     last SWITCH };
				/u/ and do { $bit = $$user{user};     last SWITCH };
				/a/ and do { $bit = $$user{addr};     last SWITCH };
				/o/ and do { $bit = $$user{prefix};   last SWITCH };
				/s/ and do { $bit = $$event{msg};     last SWITCH };
			   /w/ and do { $bit = $$event{network}; last SWITCH };
				/c/ and do { $bit = $$event{where};   last SWITCH };
				
				# bot data
				/m/i and do { $bit = get_info('nick'); last SWITCH };
				
				# macro character itself
				$_ eq MACRO_CHAR and do { $bit = MACRO_CHAR; last SWITCH };
				
				# captures: %1-%9
				/^\d$/ and do { $bit = 
					exists $$matches[$_-1] ? $$matches[$_-1] : ''; last SWITCH };
				/\$/   and do { $bit = 
					exists $$matches[$#$matches] ? $$matches[$#$matches] : ''; last SWITCH };
				
				# only do these when needed in order to save time
				/d/i and do { $bit = strftime('%Y/%m/%d', localtime); last SWITCH}; # local date
				/y/i and do { $bit = strftime('%Y/%m/%d', gmtime   ); last SWITCH}; # zulu date
				/t/i and do { $bit = strftime('%H:%M:%S', localtime); last SWITCH}; # local time
				/z/i and do { $bit = strftime('%H:%M:%S', gmtime   ); last SWITCH}; # zulu time
				
				# let color codes pass through
				/B/ and do { $bit = "\cB"; last SWITCH }; # %B == \cB
				/C/ and do { $bit = "\cC"; last SWITCH }; # %C == \cC
				/O/ and do { $bit = "\cO"; last SWITCH }; # %O == \cO
				/R/ and do { $bit = "\cV"; last SWITCH }; # %R == \cV
				/U/ and do { $bit = "\c_"; last SWITCH }; # %U == \c_
				
				# else, remove non-matching macros
				$bit = ''; 
			} # SWITCH
		} # bits
		$cmd = join '', @bits;
	} # cmds
	return @cmds;
}
sub strip_codes($) { # strip color codes for print events which laden nicks with them
	my $s = shift;
	$s =~ 
	  s!\cB|	# Bold
		\cC\d{0,2}(?:,\d{0,2})?| # Color
		\cG|	# Beep
		\cO|	# Reset
		\cV|	# Reverse
		\c_	 # Underline
	   !!gx;
	return $s;
}
sub hookage        { # get full conf path, attempt to load it, hook or unhook event
	my $quiet = shift;
	my $conffile = get_info('xchatdir');
	$conffile =~ s!\\+!/!g; # de-windows-ize it if necessary
	$conffile =~ s!/$!!; # remove ending slash to prevent doubling up, just in case
	$conffile .= '/'.CONF_FILE;

	if( $patterns = load_conf($conffile) )
	{
		if( not @privmsg_hooks )
		{
			push @privmsg_hooks, hook_print('Channel Message',		   \&hdl_chan);
			push @privmsg_hooks, hook_print('Channel Msg Hilight',	   \&hdl_chan);
			push @privmsg_hooks, hook_print('Channel Action',			\&hdl_actn);
			push @privmsg_hooks, hook_print('Channel Action Hilight',	\&hdl_actn);
			push @privmsg_hooks, hook_print('Private Message',		   \&hdl_priv);
			push @privmsg_hooks, hook_print('Private Message to Dialog', \&hdl_priv);
		}
		xprint("$PREFIX Config loaded, events hooked.") if not $quiet;
	}
	else
	{
		if( @privmsg_hooks )
		{
			unhook($_) for @privmsg_hooks; # unhook all
			@privmsg_hooks = (); # set it to empty
		}
		xprint("$PREFIX \02Error\02: Config unable to load. Events not hooked.");
	}
}

1;
__END__

Please read the accompanied readme.txt for instructions
and more information on configuration.

 author: b0at
license: public domain

changes from v003:
- added a /who handler to enable grabbing user_info from arbitrary users on the network
	this means we can get user data during private events
- cleaned lots of code up, found at least one bug that prevented private events from working right
